@@ -17,7 +17,6 @@
  *  along with this program; if not, write to the Free Software
  *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  *
- *  $Id$
  */
 
 #include "config.h"
@@ -285,6 +284,8 @@ ephy_bookmark_action_sync_icon (GtkAction *action,
 
 		gtk_image_menu_item_set_image
 			(GTK_IMAGE_MENU_ITEM (proxy), image);
+		gtk_image_menu_item_set_always_show_image (GTK_IMAGE_MENU_ITEM (proxy),
+							   TRUE);
 	}
 
 	if (pixbuf)
@@ -414,7 +415,11 @@ entry_key_press_cb (GtkEntry *entry,
 	     event->keyval == GDK_ISO_Enter) &&
 	    state == GDK_CONTROL_MASK)
 	{
+#if GTK_CHECK_VERSION (2,21,0)
+		gtk_entry_reset_im_context (entry);
+#else
 		gtk_im_context_reset (entry->im_context);
+#endif
 
 		g_signal_emit_by_name (entry, "activate");
 
@@ -462,7 +467,8 @@ drag_data_get_cb (GtkWidget          *widget,
 
 	g_return_if_fail (location != NULL);
 
-	gtk_selection_data_set (selection_data, selection_data->target, 8, (unsigned char *)location, strlen (location));
+	gtk_selection_data_set (selection_data, gtk_selection_data_get_target (selection_data),
+				8, (unsigned char *)location, strlen (location));
 }
 
 static void
@@ -592,7 +598,7 @@ connect_proxy (GtkAction *action,
 	{
 		GtkLabel *label;
 
-		label = (GtkLabel *) ((GtkBin *) proxy)->child;
+		label = GTK_LABEL (gtk_bin_get_child (GTK_BIN (proxy)));
 
 		gtk_label_set_use_underline (label, FALSE);
 		gtk_label_set_ellipsize (label, PANGO_ELLIPSIZE_END);
