@@ -48,6 +48,7 @@ struct _EphyFindToolbarPrivate
 	gulong set_focus_handler;
 	guint source_id;
 	guint find_again_source_id;
+	guint find_source_id;
 	char *find_string;
 	guint preedit_changed : 1;
 	guint prevent_activate : 1;
@@ -72,10 +73,16 @@ enum
 
 typedef enum
 {
-	EPHY_FIND_FOUND		= 0,
-	EPHY_FIND_NOTFOUND	= 1,
-	EPHY_FIND_FOUNDWRAPPED	= 2
-} EphyEmbedFindResult;
+	EPHY_FIND_RESULT_FOUND		= 0,
+	EPHY_FIND_RESULT_NOTFOUND	= 1,
+	EPHY_FIND_RESULT_FOUNDWRAPPED	= 2
+} EphyFindResult;
+
+typedef enum
+{
+	EPHY_FIND_DIRECTION_NEXT,
+	EPHY_FIND_DIRECTION_PREV
+} EphyFindDirection;
 
 static guint signals[LAST_SIGNAL];
 
@@ -92,7 +99,7 @@ scroll_lines (WebKitWebView *web_view,
         scrolled_window = GTK_SCROLLED_WINDOW (gtk_widget_get_parent (GTK_WIDGET (web_view)));
         vadj = gtk_scrolled_window_get_vadjustment (scrolled_window);
 
-        value = vadj->value + (num_lines * vadj->step_increment);
+        value = gtk_adjustment_get_value (vadj) + (num_lines * gtk_adjustment_get_step_increment (vadj));
         gtk_adjustment_set_value (vadj, value);
 }
 
@@ -107,7 +114,7 @@ scroll_pages (WebKitWebView *web_view,
         scrolled_window = GTK_SCROLLED_WINDOW (gtk_widget_get_parent (GTK_WIDGET (web_view)));
         vadj = gtk_scrolled_window_get_vadjustment (scrolled_window);
 
-        value = vadj->value + (num_pages * vadj->page_increment);
+        value = gtk_adjustment_get_value (vadj) + (num_pages * gtk_adjustment_get_page_increment (vadj));
         gtk_adjustment_set_value (vadj, value);
 }
 
@@ -130,7 +137,7 @@ set_status_notfound_cb (EphyFindToolbar *toolbar)
 
 static void
 set_status (EphyFindToolbar *toolbar,
-	    EphyEmbedFindResult result)
+	    EphyFindResult result)
 {
 	EphyFindToolbarPrivate *priv = toolbar->priv;
 	char *text = NULL;
@@ -138,10 +145,10 @@ set_status (EphyFindToolbar *toolbar,
 
 	switch (result)
 	{
-		case EPHY_FIND_FOUND:
+		case EPHY_FIND_RESULT_FOUND:
 			text = NULL;
 			break;
-		case EPHY_FIND_NOTFOUND:
+		case EPHY_FIND_RESULT_NOTFOUND:
 			{
 				text = _("Not found");
 
@@ -154,7 +161,7 @@ set_status (EphyFindToolbar *toolbar,
 				priv->source_id = g_timeout_add (500, (GSourceFunc) set_status_notfound_cb, toolbar);
 			}
 			break;
-		case EPHY_FIND_FOUNDWRAPPED:
+		case EPHY_FIND_RESULT_FOUNDWRAPPED:
 			text = _("Wrapped");
 			break;
 	}
@@ -178,7 +185,6 @@ clear_status (EphyFindToolbar *toolbar)
 			    priv->links_only ? _("Find links:") : _("Find:"));
 }
 
-#if 0
 /* Code adapted from gtktreeview.c:gtk_tree_view_key_press() and
  * gtk_tree_view_real_start_interactive_seach()
  */
@@ -187,16 +193,12 @@ tab_search_key_press_cb (EphyEmbed *embed,
 			 GdkEventKey *event,
 			 EphyFindToolbar *toolbar)
 {
-	EphyFindToolbarPrivate *priv = toolbar->priv;
 	GtkWidget *widget = (GtkWidget *) toolbar;
 
 	g_return_val_if_fail (event != NULL, FALSE);
 
-	/* don't do anything in PPV mode */
-	if (ephy_window_get_is_print_preview (priv->window)) return FALSE;
-
 	/* check for / and ' which open the find toolbar in text resp. link mode */
-	if (GTK_WIDGET_VISIBLE (widget) == FALSE)
+	if (gtk_widget_get_visible (widget) == FALSE)
 	{
 		if (event->keyval == GDK_slash)
 		{
@@ -212,7 +214,6 @@ tab_search_key_press_cb (EphyEmbed *embed,
 
 	return FALSE;
 }
-#endif
 
 static void
 find_next_cb (EphyFindToolbar *toolbar)
@@ -244,15 +245,16 @@ ephy_find_toolbar_mark_matches (EphyFindToolbar *toolbar)
         webkit_web_view_set_highlight_text_matches (web_view, TRUE);
 }
 
-static EphyEmbedFindResult
+static EphyFindResult
 real_find (EphyFindToolbarPrivate *priv,
-	   gboolean forward)
+	   EphyFindDirection direction)
 {
         WebKitWebView *web_view = priv->web_view;
         gboolean case_sensitive = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (priv->case_sensitive));
+        gboolean forward = (direction == EPHY_FIND_DIRECTION_NEXT);
 
         if (!priv->find_string || !g_strcmp0 (priv->find_string, ""))
-                return EPHY_FIND_NOTFOUND;
+                return EPHY_FIND_RESULT_NOTFOUND;
 
         if (!webkit_web_view_search_text
             (web_view, priv->find_string, case_sensitive, forward, FALSE)) {
@@ -260,14 +262,30 @@ real_find (EphyFindToolbarPrivate *priv,
                 if (!webkit_web_view_search_text
                     (web_view, priv->find_string, case_sensitive, forward, TRUE)) {
                         /* there's no result */
-                        return EPHY_FIND_NOTFOUND;
+                        return EPHY_FIND_RESULT_NOTFOUND;
                 } else {
                         /* found wrapped */
-                        return EPHY_FIND_FOUNDWRAPPED;
+                        return EPHY_FIND_RESULT_FOUNDWRAPPED;
                 }
         }
 
-        return EPHY_FIND_FOUND;
+        return EPHY_FIND_RESULT_FOUND;
+}
+
+static gboolean
+do_search (EphyFindToolbar *toolbar)
+{
+	EphyFindToolbarPrivate *priv = toolbar->priv;
+	EphyFindResult result;
+
+	priv->find_source_id = 0;
+
+	ephy_find_toolbar_mark_matches (toolbar);
+
+	result = real_find (priv, EPHY_FIND_DIRECTION_NEXT);
+	set_status (toolbar, result);
+
+	return FALSE;
 }
 
 static void
@@ -275,15 +293,19 @@ entry_changed_cb (GtkEntry *entry,
 		  EphyFindToolbar *toolbar)
 {
 	EphyFindToolbarPrivate *priv = toolbar->priv;
-	EphyEmbedFindResult result;
 
 	g_free (priv->find_string);
 	priv->find_string = g_strdup (gtk_entry_get_text (GTK_ENTRY (priv->entry)));
 
-	ephy_find_toolbar_mark_matches (toolbar);
+	if (priv->find_source_id != 0) {
+		g_source_remove (priv->find_source_id);
+		priv->find_source_id = 0;
+	}
 
-	result = real_find (priv, TRUE);
-	set_status (toolbar, result);
+	if (strlen (priv->find_string) == 0)
+                return;
+
+	priv->find_source_id = g_timeout_add (300, (GSourceFunc)do_search, toolbar);
 }
 
 static gboolean
@@ -414,11 +436,12 @@ case_sensitive_toggled_cb (GtkWidget *check,
 	 */
 	if (case_sensitive)
 	{
-		EphyEmbedFindResult result;
+		EphyFindResult result;
 
-		result = real_find (toolbar->priv, FALSE);
-		if (result != EPHY_FIND_NOTFOUND)
-			result = real_find (toolbar->priv, TRUE);
+		result = real_find (toolbar->priv, EPHY_FIND_DIRECTION_PREV);
+		if (result != EPHY_FIND_RESULT_NOTFOUND)
+			result = real_find (toolbar->priv,
+					    EPHY_FIND_DIRECTION_NEXT);
 
 		set_status (toolbar, result);
 	}
@@ -456,7 +479,7 @@ set_focus_cb (EphyWindow *window,
 
 	while (widget != NULL && widget != wtoolbar)
 	{
-		widget = widget->parent;
+		widget = gtk_widget_get_parent (widget);
 	}
 
 	/* if widget == toolbar, the new focus widget is in the toolbar */
@@ -483,7 +506,8 @@ ephy_find_toolbar_parent_set (GtkWidget *widget,
 	EphyFindToolbarPrivate *priv = toolbar->priv;
 	GtkWidget *toplevel;
 
-	if (widget->parent != NULL && priv->set_focus_handler == 0)
+	if (gtk_widget_get_parent (widget) != NULL &&
+	    priv->set_focus_handler == 0)
 	{
 		toplevel = gtk_widget_get_toplevel (widget);
 		priv->set_focus_handler =
@@ -613,6 +637,12 @@ ephy_find_toolbar_dispose (GObject *object)
 		priv->find_again_source_id = 0;
 	}
 
+	if (priv->find_source_id != 0)
+	{
+		g_source_remove (priv->find_source_id);
+		priv->find_source_id = 0;
+	}
+
 	G_OBJECT_CLASS (ephy_find_toolbar_parent_class)->dispose (object);
 }
 
@@ -737,10 +767,9 @@ ephy_find_toolbar_set_embed (EphyFindToolbar *toolbar,
 
 	if (priv->web_view != NULL)
 	{
-                /*
-		g_signal_handlers_disconnect_matched (embed, G_SIGNAL_MATCH_DATA,
-						      0, 0, NULL, NULL, toolbar);
-                */
+                g_signal_handlers_disconnect_matched (EPHY_WEB_VIEW (web_view),
+                                                      G_SIGNAL_MATCH_DATA,
+                                                      0, 0, NULL, NULL, toolbar);
 	}
 
 	priv->web_view = web_view;
@@ -748,18 +777,17 @@ ephy_find_toolbar_set_embed (EphyFindToolbar *toolbar,
 	{
 		clear_status (toolbar);
 
-		/* FIXME: this is never emitted
-		g_signal_connect_object (embed, "ge-search-key-press",
+		g_signal_connect_object (EPHY_WEB_VIEW (web_view), "search-key-press",
 					 G_CALLBACK (tab_search_key_press_cb),
 					 toolbar, 0);
-                */
 	}
 }
 
 typedef struct
 {
 	EphyFindToolbar *toolbar;
-	gboolean next;
+	gboolean direction;
+	gboolean highlight;
 } FindAgainCBStruct;
 
 static void
@@ -768,24 +796,19 @@ find_again_data_destroy_cb (FindAgainCBStruct *data)
 	g_slice_free (FindAgainCBStruct, data);
 }
 
-static EphyEmbedFindResult
-ephy_find_toolbar_find_again (EphyFindToolbar *toolbar,
-                              gboolean forward,
-                              gboolean links_only)
-{
-  EphyFindToolbarPrivate *priv = toolbar->priv;
-
-  return real_find (priv, forward);
-}
-
 static gboolean
 find_again_cb (FindAgainCBStruct *data)
 {
-	EphyEmbedFindResult result;
+	EphyFindResult result;
 	EphyFindToolbarPrivate *priv = data->toolbar->priv;
 
-	result = ephy_find_toolbar_find_again (data->toolbar, data->next,
-				  priv->links_only);
+	result = real_find (priv, data->direction);
+
+	/* Highlight matches again if the toolbar was hidden when the user
+	 * requested find-again. */
+	if (result != EPHY_FIND_RESULT_NOTFOUND && data->highlight)
+		ephy_find_toolbar_mark_matches (data->toolbar);
+
 	set_status (data->toolbar, result);
 
 	priv->find_again_source_id = 0;
@@ -793,28 +816,30 @@ find_again_cb (FindAgainCBStruct *data)
 	return FALSE;
 }
 
-void
-ephy_find_toolbar_find_next (EphyFindToolbar *toolbar)
+static void
+find_again (EphyFindToolbar *toolbar, EphyFindDirection direction)
 {
 	GtkWidget *widget = GTK_WIDGET (toolbar);
 	EphyFindToolbarPrivate *priv = toolbar->priv;
 	FindAgainCBStruct *data;
+	gboolean visible;
 
-	if (!GTK_WIDGET_VISIBLE (widget)) {
+	visible = gtk_widget_get_visible (widget);
+	if (!visible) {
 		gtk_widget_show (widget);
 		gtk_widget_grab_focus (widget);
 	}
 
-	/* We need to do this here (and in find_previous) to give time to the embed
-	 * to sync with the size change due to the toolbar being shown, otherwise
-	 * the toolbar can obscure the result. See GNOME bug #415074.
+	/* We need to do this to give time to the embed to sync with the size
+	 * change due to the toolbar being shown, otherwise the toolbar can
+	 * obscure the result. See GNOME bug #415074.
 	 */
-
 	if (priv->find_again_source_id != 0) return;
 
 	data = g_slice_new0 (FindAgainCBStruct);
 	data->toolbar = toolbar;
-	data->next = TRUE;
+	data->direction = direction;
+	data->highlight = !visible;
 
 	priv->find_again_source_id = g_idle_add_full (G_PRIORITY_DEFAULT_IDLE,
 						      (GSourceFunc) find_again_cb,
@@ -822,28 +847,16 @@ ephy_find_toolbar_find_next (EphyFindToolbar *toolbar)
 						      (GDestroyNotify) find_again_data_destroy_cb);
 }
 
+void
+ephy_find_toolbar_find_next (EphyFindToolbar *toolbar)
+{
+	find_again (toolbar, EPHY_FIND_DIRECTION_NEXT);
+}
+
 void
 ephy_find_toolbar_find_previous (EphyFindToolbar *toolbar)
 {
-	GtkWidget *widget = GTK_WIDGET (toolbar);
-	EphyFindToolbarPrivate *priv = toolbar->priv;
-	FindAgainCBStruct *data;
-
-	if (!GTK_WIDGET_VISIBLE (widget)) {
-		gtk_widget_show (widget);
-		gtk_widget_grab_focus (widget);
-	}
-
-	if (priv->find_again_source_id != 0) return;
-
-	data = g_slice_new0 (FindAgainCBStruct);
-	data->toolbar = toolbar;
-	data->next = FALSE;
-
-	priv->find_again_source_id = g_idle_add_full (G_PRIORITY_DEFAULT_IDLE,
-						      (GSourceFunc) find_again_cb,
-						      data,
-						      (GDestroyNotify) find_again_data_destroy_cb);
+	find_again (toolbar, EPHY_FIND_DIRECTION_PREV);
 }
 
 void
@@ -866,15 +879,6 @@ ephy_find_toolbar_open (EphyFindToolbar *toolbar,
 	gtk_widget_grab_focus (GTK_WIDGET (toolbar));
 }
 
-static void
-ephy_find_toolbar_set_selection (EphyFindToolbar *toolbar,
-                                 gboolean attention)
-{
-        WebKitWebView *web_view = toolbar->priv->web_view;
-
-        webkit_web_view_set_highlight_text_matches (web_view, attention);
-}
-
 void
 ephy_find_toolbar_close (EphyFindToolbar *toolbar)
 {
@@ -883,13 +887,14 @@ ephy_find_toolbar_close (EphyFindToolbar *toolbar)
 	gtk_widget_hide (GTK_WIDGET (toolbar));
 
 	if (priv->web_view == NULL) return;
-	ephy_find_toolbar_set_selection (toolbar, FALSE);
+
+	webkit_web_view_set_highlight_text_matches (priv->web_view, FALSE);
 }
 
 void
 ephy_find_toolbar_request_close (EphyFindToolbar *toolbar)
 {
-	if (GTK_WIDGET_VISIBLE (GTK_WIDGET (toolbar)))
+	if (gtk_widget_get_visible (GTK_WIDGET (toolbar)))
 	{
 		g_signal_emit (toolbar, signals[CLOSE], 0);
 	}
